/* Include files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "common.c"

/* Constants */
#define EXEFILENAME    "SS"
#define VERSIONSTRING  "Spell Searcher v6 by Morten Helles (dragon@diku.dk), September 1995"

/* Structs */
typedef struct _sinfo_t {
  int  Level;
  char Name[MAXLINELENGTH],
       School[MAXLINELENGTH],
       Sphere[MAXLINELENGTH],
       Reversible[MAXLINELENGTH],
       Range[MAXLINELENGTH],
       Component[MAXLINELENGTH],
       Duration[MAXLINELENGTH],
       CastingTime[MAXLINELENGTH],
       AreaOfEffect[MAXLINELENGTH],
       SavingThrow[MAXLINELENGTH],
       Author[MAXLINELENGTH];
} sinfo_t;

/* Function forward declarations */
void BypassIntroduction(void);
int  CheckSpellQueryInteger(int, int, int);
int  CheckSpellQueryString(char *, char *, int);
int  ConvertToInt(char *);
void DecodeParameters(int, char **);
void DisplayHelp(void);
void ExecuteQuery(void);
void GetSpellPart(char *, char *, int);
void HandleLevel(char *);
void Initialize(void);
void InsertSpellQuery(char *, char *);
void LoadNextSpell(void);
void PrintLongSpellDesc(void);
void PrintShortSpellDesc(void);
void PrintSpellDesc(int);

/* Global variables */
squery_t *SpellQuery;
sinfo_t  *CurrentSpellInfo;
FILE *SpellFile;


/*******************************************************************************
* FUNCTION : BypassIntroduction                                                *
*******************************************************************************/
void BypassIntroduction(void) {
  char SearchString[MAXLINELENGTH];
  int  EndOfIntroduction = FALSE;

  do {
    if (fgets(SearchString, MAXLINELENGTH, SpellFile) == NULL) {
      printf("ERROR : No start search marker found.\n");
      printf("HELP  : Insert a start search marker 'SS6!' just before the first\n");
      printf("        spell (or cantrip) level marker.\n");
      exit(0);
    }
    else if (strstr(SearchString, "SS6!") != NULL) {
      EndOfIntroduction = TRUE;
    }
  } while (!EndOfIntroduction);
}


/*******************************************************************************
* FUNCTION : CheckSpellQueryInteger                                            *
*******************************************************************************/
int CheckSpellQueryInteger(int LowerRange, int UpperRange, int CurrentLevel) {

  return ((CurrentLevel >= LowerRange) &&
           (CurrentLevel <= UpperRange));
}


/*******************************************************************************
* FUNCTION : CheckSpellQueryString                                             *
* REMARKS  : TypeOfQuery : 0 (and), 1 (not), 2 (or)                            *
*******************************************************************************/
int CheckSpellQueryString(char *StructPart, char *SearchString, int TypeOfQuery) {
  char Substring[30];
  int  SearchOk,
       ExitLoop = FALSE,
       Counter,
       SeperatorCounter,
       OffSet;

  /* SearchOk initial value */
  SearchOk = (TypeOfQuery == 2) ? FALSE : TRUE;

  for (Counter = 0;
       (Counter != strlen(StructPart)) && !ExitLoop;
       Counter += (OffSet + 1)) {

    /* Locate search seperator */
    for (SeperatorCounter = Counter;
         StructPart[SeperatorCounter] != '+';
         SeperatorCounter++)
      ;

    /* Copy substring */
    OffSet = SeperatorCounter - Counter;
    strncpy(Substring, &StructPart[Counter], OffSet);
    Substring[OffSet] = '\0';

    /* Check if substring */
    if (((TypeOfQuery == 0) && (strstr(SearchString, Substring) == NULL)) ||
        ((TypeOfQuery == 1) && (strstr(SearchString, Substring) != NULL)) ||
        ((TypeOfQuery == 2) && (strstr(SearchString, Substring) != NULL))) {
      SearchOk = !SearchOk;
      ExitLoop = TRUE;
    }
  }

  return (Counter == 0) || SearchOk ;
}


/*******************************************************************************
* FUNCTION : CheckSpellDescription                                             *
*******************************************************************************/
int CheckSpellDescription(char *SearchKey, int NormalQuery) {
  char SearchString[MAXLINELENGTH];
  int  SubStringFound = FALSE,
       EndOfFile = FALSE,
       ExitLevel = 0;
  fpos_t FilePos;

  if (strlen(SearchKey) != 0) { /* i.e. there IS a query */

    /* Remember file position before spell description */
    if (fgetpos(SpellFile, &FilePos) != 0) {
      printf("ERROR : Problem getting file pointer.\n");
      printf("HELP  : Check if file is corrupt.\n");
      exit(0);
    }

    /* Get blank line */
    if (fgets(SearchString, MAXLINELENGTH - 1, SpellFile) == NULL) {
      printf("ERROR : Unexpected end of file. No spell description.\n");
      printf("HELP  : Check desciption of last spell in spell file.\n");
      exit(0);
    }

    /* Search past spell description */
    do {
      if (fgets(SearchString, MAXLINELENGTH - 1, SpellFile) == NULL) {
        EndOfFile = TRUE;
      }
      else if (SearchString[0] == '') { /* New spell level marker => Exit */
        ExitLevel = 2;
      }
      else if (strlen(SearchString) < 3) { /* Two empty lines in succession => Exit */
        ExitLevel++;
      }
      else {
        ExitLevel = 0;
      }

      if (strstr(SearchString, SearchKey) != NULL) {
        SubStringFound = TRUE;
      }
    } while (!SubStringFound && (ExitLevel != 2) && !EndOfFile);

    /* Retrieve file position before spell description */
    if (fsetpos(SpellFile, &FilePos) != 0) {
      printf("ERROR : Problem setting file pointer.\n");
      printf("HELP  : Check if file is corrupt.\n");
      exit(0);
    }
  }

  return (strlen(SearchKey) == 0) || /* no query */
         ((NormalQuery && SubStringFound) || (!NormalQuery && !SubStringFound));
}

/*******************************************************************************
* FUNCTION : ConvertToInt                                                      *
*******************************************************************************/
int ConvertToInt(char *StringNum) {
  int ReturnValue;

  if (strstr(StringNum, "First") != NULL) {
    ReturnValue = 1;
  }
  else if (strstr(StringNum, "Second") != NULL) {
    ReturnValue = 2;
  }
  else if (strstr(StringNum, "Third") != NULL) {
    ReturnValue = 3;
  }
  else if (strstr(StringNum, "Fourth") != NULL) {
    ReturnValue = 4;
  }
  else if (strstr(StringNum, "Fifth") != NULL) {
    ReturnValue = 5;
  }
  else if (strstr(StringNum, "Sixth") != NULL) {
    ReturnValue = 6;
  }
  else if (strstr(StringNum, "Seventh") != NULL) {
    ReturnValue = 7;
  }
  else if (strstr(StringNum, "Eighth") != NULL) {
    ReturnValue = 8;
  }
  else if (strstr(StringNum, "Ninth") != NULL) {
    ReturnValue = 9;
  }
  else if (strstr(StringNum, "Tenth") != NULL) {
    ReturnValue = 10;
  }
  else if (strstr(StringNum, "Quest") != NULL) { /* Priest Quest spells? Let's say tenth level */
    ReturnValue = 10;
  }
  else {
    printf("ERROR : Unknown spell level or type.\n");
    printf("HELP  : Check for spelling errors in spell file.\n");
    exit(0);
  }

  return ReturnValue;
}


/*******************************************************************************
* FUNCTION : DecodeParameters                                                  *
*******************************************************************************/
void DecodeParameters(int argc, char *argv[]) {
  int ParamCount,
      Counter;

  /* Check first parameter : spellfile */
  if (argc < 2) {
    DisplayHelp();
  }
  else if ((SpellFile = fopen(argv[1], "r")) == NULL) {
    printf("ERROR : Problem opening file '%s'.\n", argv[1]);
    exit(0);
  }

  /* Check the subsequent parameters (switches and search conditions) */
  for (ParamCount = 2; ParamCount != argc; ParamCount++) {

    /* Check for switch : OPTION_DESC */
    if (strncmp(argv[ParamCount], OPTION_DESC, strlen(OPTION_DESC)) == 0) {
      SpellQuery->PrintDesc = TRUE;
    }

    /* Check for switch : OPTION_STATS */
    else if (strncmp(argv[ParamCount], OPTION_STATS, strlen(OPTION_STATS)) == 0) {
      SpellQuery->PrintStats = TRUE;
    }

    /* Check for search condition : OPTION_AREAOFEFFECT */
    else if (strncmp(argv[ParamCount], OPTION_AREAOFEFFECT, strlen(OPTION_AREAOFEFFECT)) == 0) {
      InsertSpellQuery(SpellQuery->AreaOfEffect, &argv[ParamCount][strlen(OPTION_AREAOFEFFECT)]);
    }

    /* Check for search condition : OPTION_XAREAOFEFFECT */
    else if (strncmp(argv[ParamCount], OPTION_XAREAOFEFFECT, strlen(OPTION_XAREAOFEFFECT)) == 0) {
      InsertSpellQuery(SpellQuery->XAreaOfEffect, &argv[ParamCount][strlen(OPTION_XAREAOFEFFECT)]);
    }

    /* Check for search condition : OPTION_OAREAOFEFFECT */
    else if (strncmp(argv[ParamCount], OPTION_OAREAOFEFFECT, strlen(OPTION_OAREAOFEFFECT)) == 0) {
      InsertSpellQuery(SpellQuery->OAreaOfEffect, &argv[ParamCount][strlen(OPTION_OAREAOFEFFECT)]);
    }

    /* Check for search condition : OPTION_AUTHOR */
    else if (strncmp(argv[ParamCount], OPTION_AUTHOR, strlen(OPTION_AUTHOR)) == 0) {
      InsertSpellQuery(SpellQuery->Author, &argv[ParamCount][strlen(OPTION_AUTHOR)]);
    }

    /* Check for search condition : OPTION_XAUTHOR */
    else if (strncmp(argv[ParamCount], OPTION_XAUTHOR, strlen(OPTION_XAUTHOR)) == 0) {
      InsertSpellQuery(SpellQuery->XAuthor, &argv[ParamCount][strlen(OPTION_XAUTHOR)]);
    }

    /* Check for search condition : OPTION_OAUTHOR */
    else if (strncmp(argv[ParamCount], OPTION_OAUTHOR, strlen(OPTION_OAUTHOR)) == 0) {
      InsertSpellQuery(SpellQuery->OAuthor, &argv[ParamCount][strlen(OPTION_OAUTHOR)]);
    }

    /* Check for search condition : OPTION_CASTINGTIME */
    else if (strncmp(argv[ParamCount], OPTION_CASTINGTIME, strlen(OPTION_CASTINGTIME)) == 0) {
      InsertSpellQuery(SpellQuery->CastingTime, &argv[ParamCount][strlen(OPTION_CASTINGTIME)]);
    }

    /* Check for search condition : OPTION_XCASTINGTIME */
    else if (strncmp(argv[ParamCount], OPTION_XCASTINGTIME, strlen(OPTION_XCASTINGTIME)) == 0) {
      InsertSpellQuery(SpellQuery->XCastingTime, &argv[ParamCount][strlen(OPTION_XCASTINGTIME)]);
    }

    /* Check for search condition : OPTION_OCASTINGTIME */
    else if (strncmp(argv[ParamCount], OPTION_OCASTINGTIME, strlen(OPTION_OCASTINGTIME)) == 0) {
      InsertSpellQuery(SpellQuery->OCastingTime, &argv[ParamCount][strlen(OPTION_OCASTINGTIME)]);
    }

    /* Check for search condition : OPTION_COMPONENTS */
    else if (strncmp(argv[ParamCount], OPTION_COMPONENTS, strlen(OPTION_COMPONENTS)) == 0) {
      InsertSpellQuery(SpellQuery->Components, &argv[ParamCount][strlen(OPTION_COMPONENTS)]);
    }

    /* Check for search condition : OPTION_XCOMPONENTS */
    else if (strncmp(argv[ParamCount], OPTION_XCOMPONENTS, strlen(OPTION_XCOMPONENTS)) == 0) {
      InsertSpellQuery(SpellQuery->XComponents, &argv[ParamCount][strlen(OPTION_XCOMPONENTS)]);
    }

    /* Check for search condition : OPTION_OCOMPONENTS */
    else if (strncmp(argv[ParamCount], OPTION_OCOMPONENTS, strlen(OPTION_OCOMPONENTS)) == 0) {
      InsertSpellQuery(SpellQuery->OComponents, &argv[ParamCount][strlen(OPTION_OCOMPONENTS)]);
    }

    /* Check for search condition : OPTION_DESCRIPTION */
    else if (strncmp(argv[ParamCount], OPTION_DESCRIPTION, strlen(OPTION_DESCRIPTION)) == 0) {
      InsertSpellQuery(SpellQuery->Description, &argv[ParamCount][strlen(OPTION_DESCRIPTION)]);
    }

    /* Check for search condition : OPTION_XDESCRIPTION */
    else if (strncmp(argv[ParamCount], OPTION_XDESCRIPTION, strlen(OPTION_XDESCRIPTION)) == 0) {
      InsertSpellQuery(SpellQuery->XDescription, &argv[ParamCount][strlen(OPTION_XDESCRIPTION)]);
    }

    /* Check for search condition : OPTION_DURATION */
    else if (strncmp(argv[ParamCount], OPTION_DURATION, strlen(OPTION_DURATION)) == 0) {
      InsertSpellQuery(SpellQuery->Duration, &argv[ParamCount][strlen(OPTION_DURATION)]);
    }

    /* Check for search condition : OPTION_XDURATION */
    else if (strncmp(argv[ParamCount], OPTION_XDURATION, strlen(OPTION_XDURATION)) == 0) {
      InsertSpellQuery(SpellQuery->XDuration, &argv[ParamCount][strlen(OPTION_XDURATION)]);
    }

    /* Check for search condition : OPTION_ODURATION */
    else if (strncmp(argv[ParamCount], OPTION_ODURATION, strlen(OPTION_ODURATION)) == 0) {
      InsertSpellQuery(SpellQuery->ODuration, &argv[ParamCount][strlen(OPTION_ODURATION)]);
    }

    /* Check for search condition : OPTION_LEVEL */
    else if (strncmp(argv[ParamCount], OPTION_LEVEL, strlen(OPTION_LEVEL)) == 0) {
      HandleLevel(argv[ParamCount]); /* Needs special handling */
    }

    /* Check for search condition : OPTION_NAME */
    else if (strncmp(argv[ParamCount], OPTION_NAME, strlen(OPTION_NAME)) == 0) {
      InsertSpellQuery(SpellQuery->Name, &argv[ParamCount][strlen(OPTION_NAME)]);
    }

    /* Check for search condition : OPTION_XNAME */
    else if (strncmp(argv[ParamCount], OPTION_XNAME, strlen(OPTION_XNAME)) == 0) {
      InsertSpellQuery(SpellQuery->XName, &argv[ParamCount][strlen(OPTION_XNAME)]);
    }

    /* Check for search condition : OPTION_ONAME */
    else if (strncmp(argv[ParamCount], OPTION_ONAME, strlen(OPTION_ONAME)) == 0) {
      InsertSpellQuery(SpellQuery->OName, &argv[ParamCount][strlen(OPTION_ONAME)]);
    }

    /* Check for search condition : OPTION_RANGE */
    else if (strncmp(argv[ParamCount], OPTION_RANGE, strlen(OPTION_RANGE)) == 0) {
      InsertSpellQuery(SpellQuery->Range, &argv[ParamCount][strlen(OPTION_RANGE)]);
    }

    /* Check for search condition : OPTION_XRANGE */
    else if (strncmp(argv[ParamCount], OPTION_XRANGE, strlen(OPTION_XRANGE)) == 0) {
      InsertSpellQuery(SpellQuery->XRange, &argv[ParamCount][strlen(OPTION_XRANGE)]);
    }

    /* Check for search condition : OPTION_ORANGE */
    else if (strncmp(argv[ParamCount], OPTION_ORANGE, strlen(OPTION_ORANGE)) == 0) {
      InsertSpellQuery(SpellQuery->ORange, &argv[ParamCount][strlen(OPTION_ORANGE)]);
    }

    /* Check for search condition : OPTION_SAVINGTHROW */
    else if (strncmp(argv[ParamCount], OPTION_SAVINGTHROW, strlen(OPTION_SAVINGTHROW)) == 0) {
      InsertSpellQuery(SpellQuery->SavingThrow, &argv[ParamCount][strlen(OPTION_SAVINGTHROW)]);
    }

    /* Check for search condition : OPTION_XSAVINGTHROW */
    else if (strncmp(argv[ParamCount], OPTION_XSAVINGTHROW, strlen(OPTION_XSAVINGTHROW)) == 0) {
      InsertSpellQuery(SpellQuery->XSavingThrow, &argv[ParamCount][strlen(OPTION_XSAVINGTHROW)]);
    }

    /* Check for search condition : OPTION_OSAVINGTHROW */
    else if (strncmp(argv[ParamCount], OPTION_OSAVINGTHROW, strlen(OPTION_OSAVINGTHROW)) == 0) {
      InsertSpellQuery(SpellQuery->OSavingThrow, &argv[ParamCount][strlen(OPTION_OSAVINGTHROW)]);
    }

    /* Check for search condition : OPTION_SCHOOL */
    else if (strncmp(argv[ParamCount], OPTION_SCHOOL, strlen(OPTION_SCHOOL)) == 0) {
      InsertSpellQuery(SpellQuery->School, &argv[ParamCount][strlen(OPTION_SCHOOL)]);
    }

    /* Check for search condition : OPTION_XSCHOOL */
    else if (strncmp(argv[ParamCount], OPTION_XSCHOOL, strlen(OPTION_XSCHOOL)) == 0) {
      InsertSpellQuery(SpellQuery->XSchool, &argv[ParamCount][strlen(OPTION_XSCHOOL)]);
    }

    /* Check for search condition : OPTION_OSCHOOL */
    else if (strncmp(argv[ParamCount], OPTION_OSCHOOL, strlen(OPTION_OSCHOOL)) == 0) {
      InsertSpellQuery(SpellQuery->OSchool, &argv[ParamCount][strlen(OPTION_OSCHOOL)]);
    }

    /* Check for search condition : OPTION_SPHERE */
    else if (strncmp(argv[ParamCount], OPTION_SPHERE, strlen(OPTION_SPHERE)) == 0) {
      InsertSpellQuery(SpellQuery->Sphere, &argv[ParamCount][strlen(OPTION_SPHERE)]);
    }

    /* Check for search condition : OPTION_XSPHERE */
    else if (strncmp(argv[ParamCount], OPTION_XSPHERE, strlen(OPTION_XSPHERE)) == 0) {
      InsertSpellQuery(SpellQuery->XSphere, &argv[ParamCount][strlen(OPTION_XSPHERE)]);
    }

    /* Check for search condition : OPTION_OSPHERE */
    else if (strncmp(argv[ParamCount], OPTION_OSPHERE, strlen(OPTION_OSPHERE)) == 0) {
      InsertSpellQuery(SpellQuery->OSphere, &argv[ParamCount][strlen(OPTION_OSPHERE)]);
    }

    /* Error collector */
    else {
      printf("ERROR : Invalid option '%s'.\n", argv[ParamCount]);
      printf("HELP  : Type '%s' to see the options available.\n", EXEFILENAME);
      exit(0);
    }
  }

  /* Remove '+' if NOT, AND or OR query is impossible */
  if (strlen(SpellQuery->Description)  > 0) { SpellQuery->Description[strlen(SpellQuery->Description)   - 1] = '\0'; }
  if (strlen(SpellQuery->XDescription) > 0) { SpellQuery->XDescription[strlen(SpellQuery->XDescription) - 1] = '\0'; }
}


/*******************************************************************************
* FUNCTION : DisplayHelp                                                       *
*******************************************************************************/
void DisplayHelp(void) {

  printf("\n%s\n", VERSIONSTRING);
  printf("\nUsage : %s <spellfile> { <switch> | <search condition> }\n", EXEFILENAME);
  printf("\nwhere <spellfile> is the ASCII file containing the GNSB or GNPB.");
  printf("\n      <switch> is one of the following:");
  printf("\n        %s shows spell description of each spell search hit.", OPTION_DESC);
  printf("\n        %s shows spell search statistics.", OPTION_STATS);
  printf("\n      <search condition> is one of the following:");
  printf("\n        %s<string> includes spells by this author.           (also %s, %s)", OPTION_AUTHOR, OPTION_XAUTHOR, OPTION_OAUTHOR);
  printf("\n        %s<string> includes spells that use this component.  (also %s, %s)", OPTION_COMPONENTS, OPTION_XCOMPONENTS, OPTION_OCOMPONENTS);
  printf("\n        %s<string> includes spells with this duration.       (also %s, %s)", OPTION_DURATION, OPTION_XDURATION, OPTION_ODURATION);
  printf("\n        %s<string> includes spells with this area of effect. (also %s, %s)", OPTION_AREAOFEFFECT, OPTION_XAREAOFEFFECT, OPTION_OAREAOFEFFECT);
  printf("\n        %s<string> includes spells with this description.    (also %s)", OPTION_DESCRIPTION, OPTION_XDESCRIPTION);
  printf("\n        %sX[..Y]   includes spells within this range. X, Y <- [0..10].", OPTION_LEVEL);
  printf("\n        %s<string> includes spells with this name.           (also %s, %s)", OPTION_NAME, OPTION_XNAME, OPTION_ONAME);
  printf("\n        %s<string> includes spells from this sphere.         (also %s, %s)", OPTION_SPHERE, OPTION_XSPHERE, OPTION_OSPHERE);
  printf("\n        %s<string> includes spells with this range.          (also %s, %s)", OPTION_RANGE, OPTION_XRANGE, OPTION_ORANGE);
  printf("\n        %s<string> includes spells from this school.         (also %s, %s)", OPTION_SCHOOL, OPTION_XSCHOOL, OPTION_OSCHOOL);
  printf("\n        %s<string> includes spells with this casting time.   (also %s, %s)", OPTION_CASTINGTIME, OPTION_XCASTINGTIME, OPTION_OCASTINGTIME);
  printf("\n        %s<string> includes spells with this saving throw.   (also %s, %s)\n", OPTION_SAVINGTHROW, OPTION_XSAVINGTHROW, OPTION_OSAVINGTHROW);
  printf("\nEx.: %s spells.wiz %sV+M %s0..5 %sConj+Summ %s >result", EXEFILENAME, OPTION_COMPONENTS, OPTION_LEVEL, OPTION_SCHOOL, OPTION_SCHOOL, OPTION_DESC);

  exit(0);
}


/*******************************************************************************
* FUNCTION : ExecuteQuery                                                      *
*******************************************************************************/
void ExecuteQuery(void) {
  int SearchDone     = FALSE,
      PrintTimeStats = TRUE,
      SearchMisses = 0,
      SearchHits   = 0,
      TotalSearches;
  time_t StartTime,
         DoneTime;

  /* Go! (start the stop watch) */
  if (time(&StartTime) == -1) {
    PrintTimeStats = FALSE;
  }

  do {
    LoadNextSpell();

    /* Check if there is no more spells in SpellFile */
    if (feof(SpellFile)) {
      SearchDone = TRUE;
    }

    /* Added for speed up (Spells are assumed to be sorted by spell level) */
    else if (CurrentSpellInfo->Level > SpellQuery->MaxLevel) {
      SearchDone = TRUE;
    }

    /* Check spell search conditions */
    else if (CheckSpellQueryInteger(SpellQuery->MinLevel, SpellQuery->MaxLevel, CurrentSpellInfo->Level) &&
             CheckSpellQueryString(SpellQuery->AreaOfEffect,  CurrentSpellInfo->AreaOfEffect, 0) &&
             CheckSpellQueryString(SpellQuery->XAreaOfEffect, CurrentSpellInfo->AreaOfEffect, 1) &&
             CheckSpellQueryString(SpellQuery->OAreaOfEffect, CurrentSpellInfo->AreaOfEffect, 2) &&
             CheckSpellQueryString(SpellQuery->Author,        CurrentSpellInfo->Author,       0) &&
             CheckSpellQueryString(SpellQuery->XAuthor,       CurrentSpellInfo->Author,       1) &&
             CheckSpellQueryString(SpellQuery->OAuthor,       CurrentSpellInfo->Author,       2) &&
             CheckSpellQueryString(SpellQuery->CastingTime,   CurrentSpellInfo->CastingTime,  0) &&
             CheckSpellQueryString(SpellQuery->XCastingTime,  CurrentSpellInfo->CastingTime,  1) &&
             CheckSpellQueryString(SpellQuery->OCastingTime,  CurrentSpellInfo->CastingTime,  2) &&
             CheckSpellQueryString(SpellQuery->Components,    CurrentSpellInfo->Component,    0) &&
             CheckSpellQueryString(SpellQuery->XComponents,   CurrentSpellInfo->Component,    1) &&
             CheckSpellQueryString(SpellQuery->OComponents,   CurrentSpellInfo->Component,    2) &&
             CheckSpellQueryString(SpellQuery->Duration,      CurrentSpellInfo->Duration,     0) &&
             CheckSpellQueryString(SpellQuery->XDuration,     CurrentSpellInfo->Duration,     1) &&
             CheckSpellQueryString(SpellQuery->ODuration,     CurrentSpellInfo->Duration,     2) &&
             CheckSpellQueryString(SpellQuery->Name,          CurrentSpellInfo->Name,         0) &&
             CheckSpellQueryString(SpellQuery->XName,         CurrentSpellInfo->Name,         1) &&
             CheckSpellQueryString(SpellQuery->OName,         CurrentSpellInfo->Name,         2) &&
             CheckSpellQueryString(SpellQuery->Range,         CurrentSpellInfo->Range,        0) &&
             CheckSpellQueryString(SpellQuery->XRange,        CurrentSpellInfo->Range,        1) &&
             CheckSpellQueryString(SpellQuery->ORange,        CurrentSpellInfo->Range,        2) &&
             CheckSpellQueryString(SpellQuery->SavingThrow,   CurrentSpellInfo->SavingThrow,  0) &&
             CheckSpellQueryString(SpellQuery->XSavingThrow,  CurrentSpellInfo->SavingThrow,  1) &&
             CheckSpellQueryString(SpellQuery->OSavingThrow,  CurrentSpellInfo->SavingThrow,  2) &&
             CheckSpellQueryString(SpellQuery->School,        CurrentSpellInfo->School,       0) &&
             CheckSpellQueryString(SpellQuery->XSchool,       CurrentSpellInfo->School,       1) &&
             CheckSpellQueryString(SpellQuery->OSchool,       CurrentSpellInfo->School,       2) &&
             CheckSpellQueryString(SpellQuery->Sphere,        CurrentSpellInfo->Sphere,       0) &&
             CheckSpellQueryString(SpellQuery->XSphere,       CurrentSpellInfo->Sphere,       1) &&
             CheckSpellQueryString(SpellQuery->OSphere,       CurrentSpellInfo->Sphere,       2) &&
             CheckSpellDescription(SpellQuery->Description,   TRUE) &&
             CheckSpellDescription(SpellQuery->XDescription,  FALSE)) {

      /* Print spell */
      if (SpellQuery->PrintDesc) {
        PrintLongSpellDesc();
      }
      else {
        PrintShortSpellDesc();
      }
      SearchHits++;
    }
    else {
      SearchMisses++;

      /* Bypass spell description */
      PrintSpellDesc(FALSE);
    }
  } while (!SearchDone);

  /* Stop! (stop the stop watch) */
  if (!PrintTimeStats || (time(&DoneTime) == -1)) {
    PrintTimeStats = FALSE;
  }

  /* Print statistics if option OPTION_STATS is used */
  if (SpellQuery->PrintStats) {
    TotalSearches = SearchMisses + SearchHits;
    printf("\nSpells searched   : %d", TotalSearches);
    printf("\nSpell search hits : %d", SearchHits);
    printf("\nSpell hit rate    : %.1f%%", (((float) SearchHits / (float) TotalSearches) * 100));
    if (PrintTimeStats) {
      printf("\nSearch time       : %d second(s).\n", (int) difftime(DoneTime, StartTime));
    }
  }

  /* Close SpellFile */
  if (fclose(SpellFile)) {
    printf("ERROR : Problem closing file.");
    exit(0);
  }
}

/*******************************************************************************
* FUNCTION : GetSpellPart                                                      *
*******************************************************************************/
void GetSpellPart(char *Dest, char *Marker, int MarkerIsOptional) {
  static char SearchString[MAXLINELENGTH];
  fpos_t FilePos;

  /* Remember file position if optional marker isn't there */
  if (MarkerIsOptional) {
    if (fgetpos(SpellFile, &FilePos) != 0) {
      printf("ERROR : Problem getting file pointer.\n");
      printf("HELP  : Check if file is corrupt.\n");
      exit(0);
    }
  }

  /* Get marker line(s) */
  if (fgets(SearchString, 255, SpellFile) == NULL) {
    printf("ERROR : Unexpected end of file.\n");
    printf("HELP  : '%s' marker expected. Check '%s' spell.\n", Marker, CurrentSpellInfo->Name);
    exit(0);
  }

  /* Check for marker */
  if (strstr(SearchString, Marker) != NULL) {

    /* Get line and delete '\n' */
    strcpy(Dest, SearchString);
    Dest[strlen(Dest) - 1] = '\0';
  }

  /* No marker, check if optional marker */
  else if (MarkerIsOptional) {

    /* Retrieve file position */
    if (fsetpos(SpellFile, &FilePos) != 0) {
      printf("ERROR : Problem setting file pointer.\n");
      printf("HELP  : Check if file is corrupt.\n");
      exit(0);
    }

    Dest[0] = '\0';
  }
  else {
    printf("ERROR : Parsing error.\n");
    printf("HELP  : '%s' marker expected. Check '%s' spell.\n", Marker, CurrentSpellInfo->Name);
    exit(0);
  }
}


/*******************************************************************************
* FUNCTION : HandleLevel                                                       *
*******************************************************************************/
void HandleLevel(char *Option) {
  static char BufferLevel[10];
  int  Counter1,
       Counter2,
       TempLevel;

  /* Get MinLevel */
  for (Counter1 = strlen(OPTION_LEVEL), Counter2 = 0;
       Counter1 != strlen(Option) &&
       Option[Counter1] != '.';
       Counter1++, Counter2++) {
    if (!isdigit(BufferLevel[Counter2] = Option[Counter1])) {
      printf("ERROR : Invalid char '%c' in lower range. Integer expected.\n", Option[Counter1]);
      printf("HELP  : Use an integer (0,1,..,10) in lower range, fx. '%s3..5'\n", OPTION_LEVEL);
      exit(0);
    }
  }
  BufferLevel[Counter2] = '\0';
  SpellQuery->MinLevel = atoi(BufferLevel);

  /* Check MinLevel */
  if ((SpellQuery->MinLevel < 0) || (SpellQuery->MinLevel > 10)) {
    printf("ERROR : Invalid lower range.\n");
    printf("HELP  : The valid range is 0..10\n");
    exit(0);
  }

  /* Check if range-search X..Y */
  if (Counter1 != strlen(Option)) {

    /* Check and bypass '..' */
    if (Option[++Counter1] != '.') {
      printf("ERROR : Wrong range seperator. Correct range seperator : '..'\n");
      printf("HELP  : An example is '%s3..5'", OPTION_LEVEL);
      exit(0);
    }

    /* Get MaxLevel */
    for (Counter1++, Counter2 = 0;
         Counter1 != strlen(Option);
         Counter1++, Counter2++) {
      if (!isdigit(BufferLevel[Counter2] = Option[Counter1])) {
        printf("ERROR : Invalid char '%c' in upper range. Integer expected.\n", Option[Counter1]);
        printf("HELP  : Use an integer (0,1,2...) in upper range, fx. '%s3..5'\n", OPTION_LEVEL);
        exit(0);
      }
    }
    BufferLevel[Counter2] = '\0';
    SpellQuery->MaxLevel = atoi(BufferLevel);

    /* Check MaxLevel */
    if ((SpellQuery->MaxLevel < 0) || (SpellQuery->MaxLevel > 10)) {
      printf("ERROR : Invalid upper range.\n");
      printf("HELP  : The valid range is 0..10\n");
      exit(0);
    }
  }
  else { /* One-level search */
    SpellQuery->MaxLevel = SpellQuery->MinLevel;
  }

  /* Switch if user has switched MinLevel and MaxLevel */
  if (SpellQuery->MinLevel > SpellQuery->MaxLevel) {
    printf("Switching upper and lower range values.\n");
    TempLevel = SpellQuery->MinLevel;
    SpellQuery->MinLevel = SpellQuery->MaxLevel;
    SpellQuery->MaxLevel = TempLevel;
  }
}


/*******************************************************************************
* FUNCTION : Initialize                                                        *
*******************************************************************************/
void Initialize(void) {

  /* Allocate structs */
  if ((SpellQuery = malloc(sizeof(squery_t))) == NULL) {
    printf("ERROR : Allocation failure (squery_t).\n");
    printf("HELP  : Please free %d bytes.\n", sizeof(squery_t));
    exit(0);
  }

  if ((CurrentSpellInfo = malloc(sizeof(sinfo_t))) == NULL) {
    printf("ERROR : Allocation failure (sinfo_t).\n");
    printf("HELP  : Please free %d bytes.\n", sizeof(sinfo_t));
    exit(0);
  }

  /* Initialize query (i.e. setup defaults). Function is in "common.c" */
  InitQuery(SpellQuery);

  CurrentSpellInfo->Level = 0; /* Implies that the cantrip marker can be ignored */
}


/*******************************************************************************
* FUNCTION : InsertSpellQuery                                                  *
*******************************************************************************/
void InsertSpellQuery(char *StructPart, char *StringToInsert) {

  strcat(StructPart, StringToInsert);
  strcat(StructPart, "+"); /* Search seperator */
}


/*******************************************************************************
* FUNCTION : LoadNextSpell                                                     *
* PURPOSE  : After execution CurrentSpellInfo contains all information about   *
*            the current spell in spell file. The spell file pointer is just   *
*            before the spell description.                                     *
*******************************************************************************/
void LoadNextSpell(void) {
  static char SearchString[MAXLINELENGTH];
  char *ChrPos;
  int  EndOfFile     = FALSE,
       NewSpell      = FALSE,
       NewSpellLevel = FALSE;

  /* Search for either a new spell level or a spell name */
  do {
    if (fgets(SearchString, MAXLINELENGTH - 1, SpellFile) == NULL) {
      EndOfFile = TRUE;
    }
    else if (strstr(SearchString, "Spells\n") != NULL) {
      NewSpellLevel = TRUE;
    }
    else if ((strchr(SearchString, '(') != NULL) &&
             (strchr(SearchString, ')') != NULL)) {
      NewSpell = TRUE;
    }
  } while (!EndOfFile && !NewSpellLevel && !NewSpell);

  /* New spell level marker */
  if (NewSpellLevel) {

    CurrentSpellInfo->Level = ConvertToInt(SearchString);

    /* Search spell name */
    do {
      if (fgets(SearchString, MAXLINELENGTH - 1, SpellFile) == NULL) {
        printf("ERROR : Unexpected end of file.\n");
        printf("HELP  : Check around '%s' spell.\n", CurrentSpellInfo->Name);
        printf("        This error can be ignored. (I'll continue anyway.)\n");
        EndOfFile = TRUE;
      }
      else if ((strchr(SearchString, '(') != NULL) &&
             (strchr(SearchString, ')') != NULL)) {
        NewSpell = TRUE;
      }
    } while (!EndOfFile && !NewSpell);
  }

  /* If not end of file then get spell info */
  if (NewSpell) {

    /* Get school and delete '\n' */
    if ((ChrPos = strrchr(SearchString, '(')) != NULL) {
      strcpy(CurrentSpellInfo->School, ChrPos);
      CurrentSpellInfo->School[strlen(CurrentSpellInfo->School) - 1] = '\0';
    }
    else { /* i.e. no school */
      printf("ERROR : Parsing error.\n");
      printf("HELP  : Spell must belong to a school. Check '%s' spell.\n", CurrentSpellInfo->Name);
      exit(0);
    }

    /* Get name */
    *ChrPos = '\0';
    strcpy(CurrentSpellInfo->Name, SearchString);

    /* Get 'reversible', if any (not all spells are reversible) */
    GetSpellPart(CurrentSpellInfo->Reversible, "Reversible", TRUE);

    /* Get sphere, if any (only priest spells have a sphere. Mages have a blank line). */
    GetSpellPart(CurrentSpellInfo->Sphere, "Sphere:", TRUE);
    if (strlen(CurrentSpellInfo->Sphere) == 0) {
      if (fgets(SearchString, MAXLINELENGTH - 1, SpellFile) == NULL) {
        printf("ERROR : Unexpected end of file.\n");
        printf("HELP  : Check '%s' spell.\n", CurrentSpellInfo->Name);
        exit(0);
      }
    }

    /* Get range */
    GetSpellPart(CurrentSpellInfo->Range, "Range:", FALSE);

    /* Get components (Optional because of priest Quest Level Spells) */
    GetSpellPart(CurrentSpellInfo->Component, "Components:", TRUE);

    /* Get duration */
    GetSpellPart(CurrentSpellInfo->Duration, "Duration:", FALSE);

    /* Get casting time */
    GetSpellPart(CurrentSpellInfo->CastingTime, "Casting Time:", FALSE);

    /* Get area of effect */
    GetSpellPart(CurrentSpellInfo->AreaOfEffect, "Area of Effect:", FALSE);

    /* Get saving throw */
    GetSpellPart(CurrentSpellInfo->SavingThrow, "Saving Throw:", FALSE);

    /* Get author */
    GetSpellPart(CurrentSpellInfo->Author, "Author:", FALSE);
  }
}


/*******************************************************************************
* FUNCTION : PrintLongSpellDesc                                                *
*******************************************************************************/
void PrintLongSpellDesc(void) {

  /* Print spell header */
  printf("%d %s%s\n", CurrentSpellInfo->Level, CurrentSpellInfo->Name, CurrentSpellInfo->School);
  if (strlen(CurrentSpellInfo->Reversible) != 0) { printf("%s\n", CurrentSpellInfo->Reversible); } /* optional */
  if (strlen(CurrentSpellInfo->Sphere)     != 0) { printf("%s\n", CurrentSpellInfo->Sphere);     } /* optional */
  printf("%s\n", CurrentSpellInfo->Range);
  if (strlen(CurrentSpellInfo->Component)  != 0) { printf("%s\n", CurrentSpellInfo->Component);  } /* optional */
  printf("%s\n", CurrentSpellInfo->Duration);
  printf("%s\n", CurrentSpellInfo->CastingTime);
  printf("%s\n", CurrentSpellInfo->AreaOfEffect);
  printf("%s\n", CurrentSpellInfo->SavingThrow);
  printf("%s\n", CurrentSpellInfo->Author);

  PrintSpellDesc(TRUE);
}


/*******************************************************************************
* FUNCTION : PrintShortSpellDesc                                               *
*******************************************************************************/
void PrintShortSpellDesc(void) {

  /* Compress sphere information, if any */
  if (strlen(CurrentSpellInfo->Sphere) != 0) {
    strcpy(CurrentSpellInfo->Sphere, &CurrentSpellInfo->Sphere[strlen("Sphere: ")]);
  }

  printf("%d %s%s%s\n", CurrentSpellInfo->Level, CurrentSpellInfo->Name, CurrentSpellInfo->School, CurrentSpellInfo->Sphere);
  PrintSpellDesc(FALSE);
}


/*******************************************************************************
* FUNCTION : PrintSpellDesc                                                    *
*******************************************************************************/
void PrintSpellDesc(int PrintDesc) {
  static char SearchString[MAXLINELENGTH];
  int  EndOfFile = FALSE,
       ExitLevel = 0;

  /* Get blank line */
  if (fgets(SearchString, MAXLINELENGTH - 1, SpellFile) == NULL) {
    printf("ERROR : Unexpected end of file. No spell description.\n");
    printf("HELP  : Check desciption of last spell in spell file.\n");
    exit(0);
  }

  if (PrintDesc) {
    printf("%s", SearchString);
  }

  /* Search past spell description */
  do {
    if (fgets(SearchString, MAXLINELENGTH - 1, SpellFile) == NULL) {
      EndOfFile = TRUE;
    }
    else if (SearchString[0] == '') {   /* New spell level marker => Exit */
      ExitLevel = 2;
    }
    else if (strlen(SearchString) < 3) { /* Two empty lines in succession => Exit */
      ExitLevel++;
    }
    else {
      ExitLevel = 0;
    }

    if (PrintDesc && (SearchString[0] != '')) {
      printf("%s", SearchString);
    }
  } while ((ExitLevel != 2) && !EndOfFile);
}


/*******************************************************************************
* FUNCTION : Main                                                              *
*******************************************************************************/
void main(int argc, char *argv[]) {

  Initialize();
  DecodeParameters(argc, argv);
  BypassIntroduction();
  ExecuteQuery();
}

