/* Include files */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.c"

/* Constants */
#define EXEFILENAME    "SSI"
#define VERSIONSTRING  "Spell Searcher Interface v6 by Morten Helles (dragon@diku.dk), September 1995"

/* Function forward declarations */
void AddParm(char *, char *, char *);
void DecodeParameters(int, char **);
void DisplayHelp(void);
void ExecuteQuery(void);
void Initialize(void);
int  ParameterQuery(void);
void QueryLevel(void);
void QueryString(char *, char *);
void ViewResult(void);

/* Global variables */
char Editor[30],
     SpellFile[30],
     ResultFile[30];
squery_t *SpellQuery;


/*******************************************************************************
* FUNCTION : AddParm                                                           *
*******************************************************************************/
void AddParm(char *SScommand, char *StructPart, char *Option) {

  if (StructPart == NULL) {
    strcat(SScommand, " ");
    strcat(SScommand, Option);
  }
  else if (strlen(StructPart) > 0) {
    strcat(SScommand, " \"");
    strcat(SScommand, Option);
    strcat(SScommand, StructPart);
    strcat(SScommand, "\"");
  }
}


/*******************************************************************************
* FUNCTION : DecodeParameters                                                  *
*******************************************************************************/
void DecodeParameters(int argc, char *argv[]) {

  if (argc == 1) {
    strcpy(Editor, "\0"); /* No editor */
  }
  else if ((argc == 2) && (strchr(argv[1], '?') == NULL)) {
    strcpy(Editor, argv[1]);
  }
  else {
    DisplayHelp();
  }
}


/*******************************************************************************
* FUNCTION : DisplayHelp                                                       *
*******************************************************************************/
void DisplayHelp(void) {

  printf("\n%s\n", VERSIONSTRING);
  printf("\nUsage: %s [ <editor> ]\n", EXEFILENAME);
  printf("\nwhere <editor> is name and path of your own editor.\n");
  printf("\nEx.: %s c:\\util\\e\n", EXEFILENAME);

  exit(0);
}


/*******************************************************************************
* FUNCTION : ExecuteQuery                                                      *
*******************************************************************************/
void ExecuteQuery(void) {
  char SScommand[MAXLINELENGTH],
       TempString[5];

  /* SS and spell file */
  strcpy(SScommand, "SS ");
  strcat(SScommand, SpellFile);

  /* Switches */
  if (SpellQuery->PrintDesc)  AddParm(SScommand, NULL, OPTION_DESC);
  if (SpellQuery->PrintStats) AddParm(SScommand, NULL, OPTION_STATS);

  /* Options */
  AddParm(SScommand, SpellQuery->AreaOfEffect,  OPTION_AREAOFEFFECT);
  AddParm(SScommand, SpellQuery->XAreaOfEffect, OPTION_XAREAOFEFFECT);
  AddParm(SScommand, SpellQuery->OAreaOfEffect, OPTION_OAREAOFEFFECT);
  AddParm(SScommand, SpellQuery->Author,        OPTION_AUTHOR);
  AddParm(SScommand, SpellQuery->XAuthor,       OPTION_XAUTHOR);
  AddParm(SScommand, SpellQuery->OAuthor,       OPTION_OAUTHOR);
  AddParm(SScommand, SpellQuery->CastingTime,   OPTION_CASTINGTIME);
  AddParm(SScommand, SpellQuery->XCastingTime,  OPTION_XCASTINGTIME);
  AddParm(SScommand, SpellQuery->OCastingTime,  OPTION_OCASTINGTIME);
  AddParm(SScommand, SpellQuery->Components,    OPTION_COMPONENTS);
  AddParm(SScommand, SpellQuery->XComponents,   OPTION_XCOMPONENTS);
  AddParm(SScommand, SpellQuery->OComponents,   OPTION_OCOMPONENTS);
  AddParm(SScommand, SpellQuery->Description,   OPTION_DESCRIPTION);
  AddParm(SScommand, SpellQuery->XDescription,  OPTION_XDESCRIPTION);
  AddParm(SScommand, SpellQuery->Duration,      OPTION_DURATION);
  AddParm(SScommand, SpellQuery->XDuration,     OPTION_XDURATION);
  AddParm(SScommand, SpellQuery->ODuration,     OPTION_ODURATION);
  AddParm(SScommand, SpellQuery->Name,          OPTION_NAME);
  AddParm(SScommand, SpellQuery->XName,         OPTION_XNAME);
  AddParm(SScommand, SpellQuery->OName,         OPTION_ONAME);
  AddParm(SScommand, SpellQuery->Range,         OPTION_RANGE);
  AddParm(SScommand, SpellQuery->XRange,        OPTION_XRANGE);
  AddParm(SScommand, SpellQuery->ORange,        OPTION_ORANGE);
  AddParm(SScommand, SpellQuery->SavingThrow,   OPTION_SAVINGTHROW);
  AddParm(SScommand, SpellQuery->XSavingThrow,  OPTION_XSAVINGTHROW);
  AddParm(SScommand, SpellQuery->OSavingThrow,  OPTION_OSAVINGTHROW);
  AddParm(SScommand, SpellQuery->School,        OPTION_SCHOOL);
  AddParm(SScommand, SpellQuery->XSchool,       OPTION_XSCHOOL);
  AddParm(SScommand, SpellQuery->OSchool,       OPTION_OSCHOOL);
  AddParm(SScommand, SpellQuery->Sphere,        OPTION_SPHERE);
  AddParm(SScommand, SpellQuery->XSphere,       OPTION_XSPHERE);
  AddParm(SScommand, SpellQuery->OSphere,       OPTION_OSPHERE);

  /* Level option */
  strcat(SScommand, " -l=");
  if (SpellQuery->MinLevel < 10) {
    TempString[0] = SpellQuery->MinLevel + '0';
    TempString[1] = '\0';
    strcat(SScommand, TempString);
  }
  else {
    strcat(SScommand, "10");
  }
  strcat(SScommand, "..");
  if (SpellQuery->MaxLevel < 10) {
    TempString[0] = SpellQuery->MaxLevel + '0';
    TempString[1] = '\0';
    strcat(SScommand, TempString);
  }
  else {
    strcat(SScommand, "10");
  }

  /* Result file */
  strcat(SScommand, " > ");
  strcat(SScommand, ResultFile);

  /* Print query */
  printf("\nQuery passed : %s", SScommand);
  printf("\nPlease wait ... ");
  fflush(stdout);

  /* Run the query */
  system(SScommand);

  printf("done.\n");
}


/*******************************************************************************
* FUNCTION : Initialize                                                        *
*******************************************************************************/
void Initialize(void) {

  /* Allocate structs */
  if ((SpellQuery = malloc(sizeof(squery_t))) == NULL) {
    printf("ERROR : Allocation failure (squery_t).\n");
    printf("HELP  : Please free %d bytes.\n", sizeof(squery_t));
    exit(0);
  }

  /* Initialize query (i.e. setup defaults) */
  InitQuery(SpellQuery);

  /* File name defaults */
  strcpy(ResultFile, "result");
  strcpy(SpellFile, "spells.wiz");
}


/*******************************************************************************
* FUNCTION : ParameterQuery                                                    *
*******************************************************************************/
int ParameterQuery(void) {
  char InputString[MAXLINELENGTH];
  int  RunQuery = FALSE,
       Quit     = FALSE;

  do {

    printf("\n******************************************************************");
    printf("\n********** Spell Searcher Interface v6 by Morten Helles **********");
    printf("\n******************************************************************");
    printf("\n(01) Show spell description of each search hit : %s", (SpellQuery->PrintDesc) ? "Yes" : "No");
    printf("\n(02) Show spell search statistics              : %s", (SpellQuery->PrintStats) ? "Yes" : "No");
    printf("\n(03) Spell file   : '%s'", SpellFile);
    printf("\n(04) Result file  : '%s'", ResultFile);
    printf("\n(05) Area of Eff. : '%s' (06) Not: '%s' (07) Or: '%s'", SpellQuery->AreaOfEffect, SpellQuery->XAreaOfEffect, SpellQuery->OAreaOfEffect);
    printf("\n(08) Author       : '%s' (09) Not: '%s' (10) Or: '%s'", SpellQuery->Author, SpellQuery->XAuthor, SpellQuery->OAuthor);
    printf("\n(11) Casting time : '%s' (12) Not: '%s' (13) Or: '%s'", SpellQuery->CastingTime, SpellQuery->XCastingTime, SpellQuery->OCastingTime);
    printf("\n(14) Components   : '%s' (15) Not: '%s' (16) Or: '%s'", SpellQuery->Components, SpellQuery->XComponents, SpellQuery->OComponents);
    printf("\n(17) Duration     : '%s' (18) Not: '%s' (19) Or: '%s'", SpellQuery->Duration, SpellQuery->XDuration, SpellQuery->ODuration);
    printf("\n(20) Name         : '%s' (21) Not: '%s' (22) Or: '%s'", SpellQuery->Name, SpellQuery->XName, SpellQuery->OName);
    printf("\n(23) Range        : '%s' (24) Not: '%s' (25) Or: '%s'", SpellQuery->Range, SpellQuery->XRange, SpellQuery->ORange);
    printf("\n(26) Saving throw : '%s' (27) Not: '%s' (28) Or: '%s'", SpellQuery->SavingThrow, SpellQuery->XSavingThrow, SpellQuery->OSavingThrow);
    printf("\n(29) School       : '%s' (30) Not: '%s' (31) Or: '%s'", SpellQuery->School, SpellQuery->XSchool, SpellQuery->OSchool);
    printf("\n(32) Sphere       : '%s' (33) Not: '%s' (34) Or: '%s'", SpellQuery->Sphere, SpellQuery->XSphere, SpellQuery->OSphere);
    printf("\n(35) Description  : '%s' (36) Not: '%s'", SpellQuery->Description, SpellQuery->XDescription);
    printf("\n(37) Level range  : %d to %d", SpellQuery->MinLevel, SpellQuery->MaxLevel);
    printf("\n(C) Clear query.");
    printf("\n(R) Run query.");
    printf("\n(Q) Quit.\n");
    printf("\nSelect : ");
    fflush(stdout);
    gets(InputString);

    if (toupper(InputString[0]) == 'Q') {
      Quit = TRUE;
    }
    else if (toupper(InputString[0]) == 'R') {
      RunQuery = TRUE;
    }
    else if (toupper(InputString[0]) == 'C') {
      InitQuery(SpellQuery);
    }
    else {

      switch (atoi(InputString)) {
        case 1  : SpellQuery->PrintDesc  = !SpellQuery->PrintDesc;  break;
        case 2  : SpellQuery->PrintStats = !SpellQuery->PrintStats; break;
        case 3  : QueryString(SpellFile,  "name of spell file");  break;
        case 4  : QueryString(ResultFile, "name of result file"); break;
        case 5  : QueryString(SpellQuery->AreaOfEffect,  "area of effect"); break;
        case 6  : QueryString(SpellQuery->XAreaOfEffect, "area of effect"); break;
        case 7  : QueryString(SpellQuery->OAreaOfEffect, "area of effect"); break;
        case 8  : QueryString(SpellQuery->Author,        "author");         break;
        case 9  : QueryString(SpellQuery->XAuthor,       "author");         break;
        case 10 : QueryString(SpellQuery->OAuthor,       "author");         break;
        case 11 : QueryString(SpellQuery->CastingTime,   "casting time");   break;
        case 12 : QueryString(SpellQuery->XCastingTime,  "casting time");   break;
        case 13 : QueryString(SpellQuery->OCastingTime,  "casting time");   break;
        case 14 : QueryString(SpellQuery->Components,    "components");     break;
        case 15 : QueryString(SpellQuery->XComponents,   "components");     break;
        case 16 : QueryString(SpellQuery->OComponents,   "components");     break;
        case 17 : QueryString(SpellQuery->Duration,      "duration");       break;
        case 18 : QueryString(SpellQuery->XDuration,     "duration");       break;
        case 19 : QueryString(SpellQuery->ODuration,     "duration");       break;
        case 20 : QueryString(SpellQuery->Name,          "name");           break;
        case 21 : QueryString(SpellQuery->XName,         "name");           break;
        case 22 : QueryString(SpellQuery->OName,         "name");           break;
        case 23 : QueryString(SpellQuery->Range,         "range");          break;
        case 24 : QueryString(SpellQuery->XRange,        "range");          break;
        case 25 : QueryString(SpellQuery->ORange,        "range");          break;
        case 26 : QueryString(SpellQuery->SavingThrow,   "saving throw");   break;
        case 27 : QueryString(SpellQuery->XSavingThrow,  "saving throw");   break;
        case 28 : QueryString(SpellQuery->OSavingThrow,  "saving throw");   break;
        case 29 : QueryString(SpellQuery->School,        "school");         break;
        case 30 : QueryString(SpellQuery->XSchool,       "school");         break;
        case 31 : QueryString(SpellQuery->OSchool,       "school");         break;
        case 32 : QueryString(SpellQuery->Sphere,        "sphere");         break;
        case 33 : QueryString(SpellQuery->XSphere,       "sphere");         break;
        case 34 : QueryString(SpellQuery->OSphere,       "sphere");         break;
        case 35 : QueryString(SpellQuery->Description,   "description");    break;
        case 36 : QueryString(SpellQuery->XDescription,  "description");    break;
        case 37 : QueryLevel(); break;
        default : printf("\nUnknown command!\n"); break;
      }
    }
  } while (!RunQuery && !Quit);

  return Quit;
}

/*******************************************************************************
* FUNCTION : QueryLevel                                                        *
*******************************************************************************/
void QueryLevel(void) {
  char InputString[MAXLINELENGTH];
  int  TempLevel;

  /* Get min. level */
  do {
    printf("Please enter minimum level: ");
    fflush(stdout);
    gets(InputString);
    SpellQuery->MinLevel = atoi(InputString);
  } while (!((SpellQuery->MinLevel >= 0) && (SpellQuery->MinLevel <= 10)));

  /* Get max. level */
  do {
    printf("Please enter maximum level: ");
    fflush(stdout);
    gets(InputString);
    SpellQuery->MaxLevel = atoi(InputString);
  } while (!((SpellQuery->MaxLevel >= 0) && (SpellQuery->MaxLevel <= 10)));

  /* Switch values if neccesary */
  if (SpellQuery->MinLevel > SpellQuery->MaxLevel) {
    TempLevel = SpellQuery->MaxLevel;
    SpellQuery->MaxLevel = SpellQuery->MinLevel;
    SpellQuery->MinLevel = TempLevel;
  }
}


/*******************************************************************************
* FUNCTION : QueryString                                                       *
*******************************************************************************/
void QueryString(char *StructPart, char *AskPart) {

  printf("\nPlease enter %s: ", AskPart);
  fflush(stdout);
  gets(StructPart);
}


/*******************************************************************************
* FUNCTION : ViewResult                                                        *
*******************************************************************************/
void ViewResult(void) {
  char OScommand[MAXLINELENGTH],
       InputString[MAXLINELENGTH],
       InputValue = 0;
  int  Counter,
       QuitListing = FALSE;
  FILE *ResultStream;

  if (strlen(Editor) == 0) { /* No editor */

    /* Ask if user wants to view the search results from here */
    printf("\nYou specified no editor as a parameter to %s.", EXEFILENAME);
    printf("\nDo you wish to view the result from here (y/n) ? ");
    fflush(stdout);
    gets(InputString);

    if (toupper(InputString[0]) == 'Y') {

      /* Get number of lines pr. screen */
      do {
        printf("\nPlease write number of lines on your screen: ");
        fflush(stdout);
        gets(InputString);
        InputValue = atoi(InputString);
      } while (InputValue < 1);

      /* Open result file */
      if ((ResultStream = fopen(ResultFile, "r")) == NULL) {
        printf("ERROR : Problem opening file '%s'.\n", ResultFile);
        exit(0);
      }

      /* Print result on screen */
      while (!feof(ResultStream) && !QuitListing) {

        /* Print one screen */
        fgets(InputString, MAXLINELENGTH, ResultStream);
        for (Counter = 0;
             (Counter != InputValue) && (!feof(ResultStream));
             Counter++, fgets(InputString, MAXLINELENGTH, ResultStream)) {
          printf("%s", InputString);
        }

        /* Wait */
        printf("Press return to continue or 'q' to quit.");
        fflush(stdout);
        gets(InputString);
        if (toupper(InputString[0]) == 'Q') {
          QuitListing = TRUE;
        }
      }

      /* Close result file */
      if (fclose(ResultStream)) {
        printf("ERROR : Problem closing file '%s'.\n", ResultFile);
        exit(0);
      }
    }
  }
  else { /* Editor specified as a parameter */
    printf("Please wait while invoking editor ...\n");
    strcpy(OScommand, Editor);
    strcat(OScommand, " ");
    strcat(OScommand, ResultFile);
    system(OScommand);
  }
}


/*******************************************************************************
* FUNCTION : Main                                                              *
*******************************************************************************/
void main(int argc, char *argv[]) {
  int Quit;

  Initialize();
  DecodeParameters(argc, argv);

  do {
    Quit = ParameterQuery();
    if (!Quit) {
      ExecuteQuery();
      ViewResult();
    }
  } while (!Quit);
}
